@@ -1,20 +1,20 @@
 """Declares request handlers for liveness and readyness
 checks.
 """
-from fastapi.responses import Response
+from .nocacheresponse import NoCacheResponse
 
 
-async def ready(response: Response) -> Response:
+async def ready() -> NoCacheResponse:
     """Perform a readyness check and set status code ``204``
     on the HTTP response if the application is ready,
     else ``503``.
     """
-    response.status_code = 204
+    return NoCacheResponse(content={'status': "OK"})
 
 
-async def live(response: Response) -> Response:
+async def live() -> NoCacheResponse:
     """Perform a liveness check and set status code ``204``
     on the HTTP response if the application is live,
     else ``503``.
     """
-    response.status_code = 204
+    return NoCacheResponse(content={'status': "OK"})
