@@ -1,5 +1,6 @@
 """Declares :class:`KeyTrustPolicy`."""
 import functools
+import logging
 import operator
 
 from unimatrix.ext.crypto import policy
@@ -10,6 +11,7 @@ from .exceptions import TrustIssues
 
 class KeyTrustPolicy:
     """Declares a policy to determine if a certain public key is trusted."""
+    logger = logging.getLogger('uvicorn')
 
     def __init__(self, tags: list, **annotations):
         if annotations:
