@@ -1,25 +1,36 @@
 # pylint: skip-file
 """Declares :class:`ResourceEndpointSet`."""
+import copy
 import functools
 import inspect
 import re
 import types
 import typing
+import urllib.parse
 from collections import OrderedDict
 
 import ioc
 from fastapi import APIRouter
 from fastapi import Depends
 from fastapi import Request
+from fastapi.responses import Response
 from fastapi.security import HTTPBearer
 from fastapi.security import HTTPAuthorizationCredentials
 from pydantic import BaseModel
 from unimatrix.conf import settings
+from unimatrix.lib.datastructures import ImmutableDTO, DTO
 
 from .auth import IHTTPAuthenticationService
+from .auth import IPrincipal
+from .dependency import inject
 from .exceptions import BearerAuthenticationRequired
+from .endpointiam import EndpointIAM
+from .exceptions import NotAuthorized
 from .exceptions import TrustIssues
+from .requestprincipalresolver import RequestPrincipalResolver
 from .resourceschema import ResourceSchema
+from .utils import clean_signature
+from .utils import clone_signature
 
 
 ACTION_METHODS = {
@@ -33,6 +44,11 @@ ACTION_METHODS = {
     'update'    : 'POST',
 }
 
+DEFAULT_ACTIONS = [
+    'create', 'retrieve', 'update', 'destroy', 'index', 'apply', 'replace',
+    'purge'
+]
+
 DETAIL_ACTIONS = {'retrieve', 'update', 'replace', 'destroy', 'apply'}
 
 
@@ -49,9 +65,44 @@ class ResourceEndpointSetMetaclass(type):
         super_new = super().__new__
         if name in ('ResourceEndpointSet', 'PublicResourceEndpointSet'):
             return super_new(cls, name, bases, attrs)
+
+        # Instantiate the IAM handler for the endpoint.
+        iam_class = attrs.pop('iam_class', EndpointIAM)
+        attrs['iam'] = iam_class(
+            issuers=set(attrs.pop('trusted_issuers', [])),
+            audiences=set(attrs.pop('accepted_audiences', [])),
+            scope=set(attrs.pop('required_scope', [])),
+            local=attrs.pop('trust_local', False)
+        )
+
         new_class = super_new(cls, name, bases, attrs)
         hints = typing.get_type_hints(new_class)
 
+        # Add the action attribute to known actions so that we can process them
+        # later on consistently with user-defined actions.
+        #
+        # TODO: Do not add them as action attribute because it confuses other
+        # methods. Refactor the *class* actions attribute to hold all actions.
+        new_class.actions = []
+        for attname in new_class.__dict__:
+            handler = new_class.__dict__[attname]
+            if attname not in DEFAULT_ACTIONS\
+            and not hasattr(handler, 'action'):
+                continue
+            if hasattr(handler, 'action'):
+                new_class.actions.append(handler.action)
+                continue
+
+            is_detail = attname in DETAIL_ACTIONS
+            action = DTO(
+                name=attname,
+                methods=[ ACTION_METHODS[attname] ],
+                detail=is_detail,
+                is_default=True,
+                path=f'{new_class.path_parameter}' if is_detail else ''
+            )
+            new_class.actions.append(action)
+
         # Convert the path_parameter attribute to an inspect.Parameter
         # instance. This is later used for constructing the appropriate
         # signature. If there is no annotation, assume that it is a string.
@@ -62,18 +113,51 @@ class ResourceEndpointSetMetaclass(type):
             default=new_class.path_parameter
         )
 
+        #: If the class has subresources, copy them and add this class as
+        #: the parent attribute.
+        new_class.subresources = copy.deepcopy(new_class.subresources or [])
+        for sub in new_class.subresources:
+            sub.parent_class = new_class
+
+            # Add the subresource itself as an action. Some special handling
+            # is required here because the class itself is not a valid named
+            # URL, so we have to do some hacking.
+            new_class.actions.append(
+                ImmutableDTO({
+                    'detail': True,
+                    'name': f'{sub.resource_name}',
+                    'path': sub.resource_name,
+                    'is_default': False,
+                    'subresource': True
+                })
+            )
+
+            # Also iterate over the actions included on the subresource
+            # class, and add them as detail actions on the parent class
+            # if they are *not* detail actions on the child class.
+            for action in sub.actions:
+                if action.detail or getattr(action, 'is_default', False):
+                    continue
+                new_class.actions.append(
+                    ImmutableDTO({
+                        'detail': True,
+                        'name': f'{sub.resource_name}.{action.name}',
+                        'path': f'{sub.resource_name}/{action.path}',
+                        'is_default': False
+                    })
+                )
+
         return new_class
 
+
 class ResourceEndpointSet(metaclass=ResourceEndpointSetMetaclass):
     """Groups a set of endpoints that allow reading, mutating and
     destroying a specific resource.
     """
+    parent_class = None
 
     #: The list of valid action names.
-    valid_actions: list = [
-        'create', 'retrieve', 'update', 'destroy', 'index', 'apply', 'replace',
-        'purge'
-    ]
+    valid_actions: list = DEFAULT_ACTIONS
 
     #: The current request that is being handled. This attribute is ``None``
     #: when initializing :class:`ResourceEndpointSet` and is set when the ASGI
@@ -86,6 +170,10 @@ class ResourceEndpointSet(metaclass=ResourceEndpointSetMetaclass):
     #: The name of the resource identifier as a path path_parameter.
     path_parameter: str = 'resource_id'
 
+    #: Set this to ``True`` if the path parameter may have
+    #: slashes.
+    path_allows_slashes: bool = False
+
     #: A list of :class:`ResourceEndpointSet` implementations representing
     #: subresources. Subresources always operate on a single object.
     subresources: list = []
@@ -123,6 +211,20 @@ class ResourceEndpointSet(metaclass=ResourceEndpointSetMetaclass):
     #: endpoint.
     trust_local: bool = False
 
+    #: Indicates if the resource is a singleton and detail-specific
+    #: behavior must be forced.
+    singleton: bool = False
+
+    #: The maximum number of results for a list request.
+    max_items: int = 100
+
+    #: The resource class used by the standard endpoints. See also
+    #: :func:`unimatrix.ext.webapi.resource`.
+    resource_class = None
+
+    #: The class used to handle IAM.
+    iam_class: type = EndpointIAM
+
     def __init__(self, action, *args, **kwargs):
         self.action = action
 
@@ -140,7 +242,9 @@ class ResourceEndpointSet(metaclass=ResourceEndpointSetMetaclass):
         `base_path`.
         """
         path_parameters = path_parameters or []
-        path_parameters.append(cls.path_parameter)
+        if not cls.singleton:
+            path_parameters.append(cls.path_parameter)
+        cls.path_parameters = copy.deepcopy(path_parameters)
         url_params = []
         base_name = str.lower(
             base_name
@@ -169,6 +273,9 @@ class ResourceEndpointSet(metaclass=ResourceEndpointSetMetaclass):
             path = base_path
             if obj.action.detail:
                 path = f'{base_path}/{{{cls.path_parameter.default}}}'
+                if cls.path_allows_slashes:
+                    path = f'{base_path}/{{{cls.path_parameter.default}:path}}'
+                path = re.sub('^[/]+', '/', path) # TODO: hack
 
                 # Check if the path parameter name is not the same as the
                 # parent class
@@ -189,7 +296,8 @@ class ResourceEndpointSet(metaclass=ResourceEndpointSetMetaclass):
                 cls._create_annotated_handler(
                     cls, obj.action.name, obj, parent=parent,
                     detail=obj.action.detail,
-                    path_parameters=path_parameters
+                    path_parameters=path_parameters,
+                    logger=router.logger
                 ),
                 name=name,
                 summary=name,
@@ -199,10 +307,12 @@ class ResourceEndpointSet(metaclass=ResourceEndpointSetMetaclass):
                     obj,
                     'response_model',
                     get_return_type(obj)),
+                response_model_exclude_defaults=True,
+                response_model_exclude_none=True,
             )
 
         has_detail_methods = False
-        has_collection_methods = False
+        has_collection_methods = False or cls.singleton
         for action in cls.valid_actions:
             # Create a fake signature to trick FastAPI in registering the
             # correct path parameters, authentication schemes, etc.
@@ -213,10 +323,13 @@ class ResourceEndpointSet(metaclass=ResourceEndpointSetMetaclass):
 
             # Add the path parameter for actions that operate on a specific
             # resource.
-            is_detail = action in DETAIL_ACTIONS
-            if is_detail:
+            is_detail = (action in DETAIL_ACTIONS)
+            if is_detail and not cls.singleton:
                 has_detail_methods = True
                 path = f'{base_path}/{{{cls.path_parameter.default}}}'
+                if cls.path_allows_slashes:
+                    path = f'{base_path}/{{{cls.path_parameter.default}:path}}'
+                path = re.sub('^[/]+', '/', path) # TODO: hack
 
                 # Check if the path parameter name is not the same as the
                 # parent class
@@ -240,7 +353,8 @@ class ResourceEndpointSet(metaclass=ResourceEndpointSetMetaclass):
                     cls, action, handler,
                     parent=parent,
                     detail=is_detail,
-                    path_parameters=path_parameters
+                    path_parameters=path_parameters,
+                    logger=router.logger
                 ),
                 name=name,
                 summary=summary,
@@ -250,6 +364,8 @@ class ResourceEndpointSet(metaclass=ResourceEndpointSetMetaclass):
                     handler,
                     'response_model',
                     get_return_type(handler)),
+                response_model_exclude_defaults=True,
+                response_model_exclude_none=True,
             )
 
         # Iterate over subresources and add them add detail endpoints.
@@ -260,7 +376,7 @@ class ResourceEndpointSet(metaclass=ResourceEndpointSetMetaclass):
                 url_params=url_params,
                 group_name=group_name,
                 base_name=base_name,
-                path_parameters=path_parameters
+                path_parameters=copy.deepcopy(path_parameters)
             )
 
         # Create the OPTIONS methods. Basically, we need to figure out what
@@ -304,7 +420,7 @@ class ResourceEndpointSet(metaclass=ResourceEndpointSetMetaclass):
 
 
     @staticmethod
-    def _update_signature(action, handler, protected=False, detail=False, path_parameters=None): # pylint: disable=line-too-long
+    def _update_signature(view_class, action, handler, protected=False, detail=False, path_parameters=None, singleton=False): # pylint: disable=line-too-long
         # Create an ordered dictionary containing the handlers' signature
         # parameters and add in the following order:
         #
@@ -338,14 +454,6 @@ class ResourceEndpointSet(metaclass=ResourceEndpointSetMetaclass):
                 break
             annotations[varname] = param
 
-        # Add the bearer token as a keyword parameter for protected endpoints.
-        if protected:
-            annotations['_bearer'] = inspect.Parameter(
-                '_bearer',
-                inspect.Parameter.POSITIONAL_OR_KEYWORD,
-                default=Depends(HTTPBearer(auto_error=False)),
-            )
-
         # Remove the self parameter (causes incorrect rendering of the UI)
         # and add the variable positional and keyword arguments, if any. This
         # should not happen because request handlers do no accept these kind
@@ -357,6 +465,31 @@ class ResourceEndpointSet(metaclass=ResourceEndpointSetMetaclass):
                 continue
             annotations[varname] = param
 
+        # Setup authentication by injecting the Endpoint.get_principal()
+        # and Endpoint.authorize() results.
+        async def authorize(*args, **kwargs):
+            return await view_class.authorize(*args, **kwargs)
+        clone_signature(view_class.authorize, authorize, {
+            'principal': inspect.Parameter(
+                name='principal',
+                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
+                default=Depends(view_class.get_principal)
+            )
+        })
+
+        annotations.update({
+            '_authorized': inspect.Parameter(
+                '_authorized',
+                inspect.Parameter.POSITIONAL_OR_KEYWORD,
+                default=Depends(authorize)
+            ),
+            '_principal': inspect.Parameter(
+                '_principal',
+                inspect.Parameter.POSITIONAL_OR_KEYWORD,
+                default=Depends(view_class.get_principal)
+            ),
+        })
+
         # Update the signature to hold the new annotations. Remove the self
         # parameter to not confuse FastAPI
         annotations.pop('self', None)
@@ -369,7 +502,7 @@ class ResourceEndpointSet(metaclass=ResourceEndpointSetMetaclass):
         return signature, annotations
 
     @staticmethod # pylint: disable=line-too-long
-    def _create_annotated_handler(view_class, action, handler, parent, detail, path_parameters):
+    def _create_annotated_handler(view_class, action, handler, parent, detail, path_parameters, logger=None):
         @functools.wraps(handler)
         async def request_handler(
             request: Request,
@@ -380,77 +513,139 @@ class ResourceEndpointSet(metaclass=ResourceEndpointSetMetaclass):
             """
             view = view_class(action)
             view.request = request
+            view.logger = logger
+            view.principal = kwargs.pop('_principal', None)
+            view.is_authorized = kwargs.pop('_authorized', False)
             return await view.dispatch(*args, **kwargs)
 
         request_handler.__signature__, request_handler.__annotations__ = (
-            view_class._update_signature(action, handler,
+            view_class._update_signature(view_class, action, handler,
                 protected=view_class.require_authentication,
                 detail=detail,
-                path_parameters=path_parameters
+                path_parameters=path_parameters,
+                singleton=view_class.singleton
             )
         )
         return request_handler
 
-    def get_audience(self) -> set:
-        """Return the audience for this endpoint that is used to verify
-        the ``aud`` claim of bearer tokens. The default implementation returns
-        an empty list.
-        """
-        return self.accepted_audiences\
-            + [f"https://{self.request.url.netloc}", "self"]
-
-    def get_scope(self) -> set:
-        """Return the scope that is required for authenticated requests."""
-        return self.required_scope
-
-    def get_trusted_issuers(self) -> set:
-        """Return the list of trusted issuers as compared against the ``iss``
-        claim.
-        """
-        return list(self.trusted_issuers)\
-            + list(settings.OAUTH2_TRUSTED_STS)\
-            + [f"https://{self.request.url.netloc}", "self"]
-
     async def dispatch(self, *args, **kwargs) -> dict:
         """Dispatch the incoming HTTP request to the appropriate request
         handler.
         """
-        if self.require_authentication:
-            await self.authenticate(kwargs.pop('_bearer', None))
+        if self.principal is None and self.require_authentication:
+            raise BearerAuthenticationRequired
+        if not self.is_authorized:
+            raise NotAuthorized
         handler = getattr(self, self.action)
-
         assert callable(handler), handler #nosec
         return await handler(*args, **kwargs)
 
-
-    @ioc.inject('authentication', 'HTTPAuthenticationService')
-    async def authenticate(self,
-        dto: HTTPAuthorizationCredentials,
-        authentication: IHTTPAuthenticationService
-    ):
-        """Authenticates the request using the bearer token provided in the
-        request.
-
-        Pass the request instance and the bearer to all authenticators and
-        return the result of the first succesful attempt. If no authenticator
-        could resolve a principal and subject, raise an exception if the
-        resource requires authentication.
+    @classmethod
+    async def get_principal(cls,
+        request: Request,
+        bearer: HTTPAuthorizationCredentials = Depends(HTTPBearer(auto_error=False)),
+        resolver = inject(
+            'RequestPrincipalResolver',
+            default=RequestPrincipalResolver()
+        )
+    ) -> IPrincipal:
+        """Return a :class:`unimatrix.ext.webapi.auth.IPrincipal` implementation
+        based on the current request and supplied credentials, or ``None`` if
+        the credentials were absent or did not resolve to a principal.
         """
-        self.principal = None
-        if dto is not None:
-            if not self.trust_policy\
-            and self.require_authentication\
-            and not self.trust_local:
-                raise TrustIssues
-            self.principal = await authentication.resolve(
-                str.encode(dto.credentials),
-                audience=self.get_audience(),
-                issuers=self.get_trusted_issuers(),
-                scope=self.get_scope(),
-                policy=self.trust_policy
-            )
-        if self.principal is None and self.require_authentication:
-            raise BearerAuthenticationRequired
+        if bearer is None:
+            return None
+        return await resolver.resolve(
+            request=request,
+            iam=cls.iam,
+            bearer=bearer.credentials,
+            policy=cls.trust_policy
+        )
+
+    @staticmethod
+    async def authorize(principal: IPrincipal, *args, **kwargs) -> bool:
+        """Hook to perform authorization. Implementations are expected to raise
+        an exception if the request is not authorized to perform the operation.
+        """
+        return True
+
+    def render_to_response(self, *args, **kwargs) -> Response:
+        """Renders a resource to a HTTP response."""
+        return self.resource_class.render_to_response(
+            self, self.request, *args, **kwargs
+        )
+
+    def reverse(self, name: str, **params) -> str:
+        """Reverses the given path relative to the current view."""
+        return self.request.url_for(f'{self.resource_name}.{name}', **params)
+
+    def get_detail_url(self, *args, **params) -> str:
+        """Returns the detail view URL of the endpoint."""
+        path_parameter = self.path_parameter.default
+        if args:
+            dto = args[0]
+            params = {path_parameter: dto[path_parameter]}
+        return self.reverse('retrieve', **params)
+
+    def get_limit(self) -> int:
+        """Return the limit for pagination from a request."""
+        limit = self._get_int_query_param(self.request.query_params, 'limit', 100)
+        return min(limit, self.max_items)
+
+    def get_next_url(self, total: int = None) -> str:
+        """Return the next URL for pagination, based on the current offset
+        and limit.
+        """
+        limit = self.get_limit()
+        params = {
+            **self.request.query_params,
+            'offset': self.get_offset() + limit,
+            'limit': limit
+        }
+
+        return f"{self.reverse('index')}?{urllib.parse.urlencode(params, doseq=False)}"\
+            if not (total is not None and params['offset'] >= total) else None
+
+    def get_offset(self) -> int:
+        """Return the offset for pagination from a request."""
+        return self._get_int_query_param(self.request.query_params, 'offset', 0)
+
+    def get_prev_url(self) -> str:
+        """Return the previous URL for pagination, based on the current offset
+        and limit.
+        """
+        limit = self.get_limit()
+        params = {
+            **self.request.query_params,
+            'offset': self.get_offset() - limit,
+            'limit': limit
+        }
+        return f"{self.reverse('index')}?{urllib.parse.urlencode(params, doseq=False)}"\
+            if params['offset'] >= 0 else None
+
+    def get_resource_links(self, resource) -> dict:
+        """Return a dictionary containing the links for a single resource."""
+        params = self._get_path_params_from_resource(resource)
+        links = {}
+        for action in self.actions:
+            if not action.detail:
+                continue
+            if not getattr(action, 'subresource', False):
+                links[action.name] = self.reverse(action.name, **params)
+            else:
+                # For subresources, the default actions are not added
+                # separately, but instead the subresource is referenced
+                # as a single link.
+                links[action.name] = self.reverse('retrieve', **params)\
+                    + '/' + action.path
+        return links
+
+    def _get_int_query_param(self, params, name, default):
+        value = params.get(name) or ''
+        return int(value if str.isdigit(value) else default)
+
+    def _get_path_params_from_resource(self, resource) -> dict:
+        return {x.default: resource[x.default] for x in self.path_parameters}
 
 
 class PublicResourceEndpointSet(ResourceEndpointSet):
